@@ -292,7 +292,7 @@ protected:
       case Call::eid:
         switch (static_cast<Call::CallKind>(n->_secondaryId)) {
           case Call::CK_NULLARY:
-            ns = sizeof(Call0);
+            ns = sizeof(Call1);
             break;
           case Call::CK_BINARY:
           case Call::CK_NARY_2:
@@ -366,6 +366,12 @@ void GC::lock() {
   }
   gc()->_lockCount++;
 }
+void GC::lockNoGC() {
+  if (gc() == nullptr) {
+    gc() = new GC();
+  }
+  gc()->_lockCount++;
+}
 void GC::unlock() {
   assert(locked());
   gc()->_lockCount--;
@@ -554,7 +560,7 @@ void GC::Heap::sweep() {
           default:
             if (n->_id >= static_cast<unsigned int>(ASTNode::NID_END) + 1 &&
                 n->_id <= static_cast<unsigned int>(Expression::EID_END)) {
-              static_cast<Expression*>(n)->ann().~Annotation();
+              Expression::ann(static_cast<Expression*>(n)).~Annotation();
             }
         }
         if (ns >= _fl_size[0] && ns <= _fl_size[_max_fl]) {
@@ -710,30 +716,30 @@ void GC::removeKeepAlive(KeepAlive* e) {
 }
 
 KeepAlive::KeepAlive(Expression* e) : _e(e), _p(nullptr), _n(nullptr) {
-  if ((_e != nullptr) && !_e->isUnboxedVal()) {
-    GC::gc()->addKeepAlive(this);
+  if ((_e != nullptr) && !Expression::isUnboxedVal(_e)) {
+    GC::addKeepAlive(this);
   }
 }
 KeepAlive::~KeepAlive() {
-  if ((_e != nullptr) && !_e->isUnboxedVal()) {
-    GC::gc()->removeKeepAlive(this);
+  if ((_e != nullptr) && !Expression::isUnboxedVal(_e)) {
+    GC::removeKeepAlive(this);
   }
 }
 KeepAlive::KeepAlive(const KeepAlive& e) : _e(e._e), _p(nullptr), _n(nullptr) {
-  if ((_e != nullptr) && !_e->isUnboxedVal()) {
-    GC::gc()->addKeepAlive(this);
+  if ((_e != nullptr) && !Expression::isUnboxedVal(_e)) {
+    GC::addKeepAlive(this);
   }
 }
 KeepAlive& KeepAlive::operator=(const KeepAlive& e) {
   if (this != &e) {
-    if ((_e != nullptr) && !_e->isUnboxedVal()) {
-      if (e._e == nullptr || e._e->isUnboxedVal()) {
-        GC::gc()->removeKeepAlive(this);
+    if ((_e != nullptr) && !Expression::isUnboxedVal(_e)) {
+      if (e._e == nullptr || Expression::isUnboxedVal(e._e)) {
+        GC::removeKeepAlive(this);
         _p = _n = nullptr;
       }
     } else {
-      if (e._e != nullptr && !e._e->isUnboxedVal()) {
-        GC::gc()->addKeepAlive(this);
+      if (e._e != nullptr && !Expression::isUnboxedVal(e._e)) {
+        GC::addKeepAlive(this);
       }
     }
     _e = e._e;
@@ -783,50 +789,52 @@ void GC::removeNodeWeakMap(ASTNodeWeakMap* m) {
 }
 
 WeakRef::WeakRef(Expression* e) : _e(e), _p(nullptr), _n(nullptr) {
-  if (_e != nullptr && !_e->isUnboxedVal()) {
-    GC::gc()->addWeakRef(this);
+  if (_e != nullptr && !Expression::isUnboxedVal(_e)) {
+    GC::addWeakRef(this);
   }
 }
 WeakRef::~WeakRef() {
-  if ((_e != nullptr) && !_e->isUnboxedVal()) {
-    GC::gc()->removeWeakRef(this);
+  if ((_e != nullptr) && !Expression::isUnboxedVal(_e)) {
+    GC::removeWeakRef(this);
   }
 }
 WeakRef::WeakRef(const WeakRef& e) : _e(e()), _p(nullptr), _n(nullptr) {
-  if (_e != nullptr && !_e->isUnboxedVal()) {
-    GC::gc()->addWeakRef(this);
+  if (_e != nullptr && !Expression::isUnboxedVal(_e)) {
+    GC::addWeakRef(this);
   }
 }
 WeakRef& WeakRef::operator=(const WeakRef& e) {
   if (this != &e) {
     // Test if this WeakRef is currently active in the GC
-    bool isActive = ((_e != nullptr) && !_e->isUnboxedVal());
+    bool isActive = ((_e != nullptr) && !Expression::isUnboxedVal(_e));
     if (isActive) {
       // Yes, active WeakRef.
       // If after assigning WeakRef should be inactive, remove it.
-      if (e() == nullptr || e()->isUnboxedVal()) {
-        GC::gc()->removeWeakRef(this);
+      if (e() == nullptr || Expression::isUnboxedVal(e())) {
+        GC::removeWeakRef(this);
         _n = _p = nullptr;
       }
     }
     _e = e();
 
     // If this WeakRef was not active but now should be, add it
-    if (!isActive && _e != nullptr && !_e->isUnboxedVal()) {
-      GC::gc()->addWeakRef(this);
+    if (!isActive && _e != nullptr && !Expression::isUnboxedVal(_e)) {
+      GC::addWeakRef(this);
     }
   }
   return *this;
 }
 
-Expression* WeakRef::operator()() { return _e->isUnboxedVal() ? _e : _p != nullptr ? _e : nullptr; }
+Expression* WeakRef::operator()() {
+  return Expression::isUnboxedVal(_e) ? _e : _p != nullptr ? _e : nullptr;
+}
 Expression* WeakRef::operator()() const {
-  return _e->isUnboxedVal() ? _e : _p != nullptr ? _e : nullptr;
+  return Expression::isUnboxedVal(_e) ? _e : _p != nullptr ? _e : nullptr;
 }
 
-ASTNodeWeakMap::ASTNodeWeakMap() : _p(nullptr), _n(nullptr) { GC::gc()->addNodeWeakMap(this); }
+ASTNodeWeakMap::ASTNodeWeakMap() : _p(nullptr), _n(nullptr) { GC::addNodeWeakMap(this); }
 
-ASTNodeWeakMap::~ASTNodeWeakMap() { GC::gc()->removeNodeWeakMap(this); }
+ASTNodeWeakMap::~ASTNodeWeakMap() { GC::removeNodeWeakMap(this); }
 
 void ASTNodeWeakMap::insert(ASTNode* n0, ASTNode* n1) { _m.insert(std::make_pair(n0, n1)); }
 
